import { isArray } from "@thi.ng/checks/is-array";
import { DGraph } from "@thi.ng/dgraph";
import { isMat, isTerm, isVec } from "./checks.js";
/**
 * Helper function for {@link walk}. Returns child nodes for any control
 * flow nodes containing a child scope.
 *
 * {@link allChildren}
 */
export const scopedChildren = (t) => t.tag === "fn" || t.tag === "for" || t.tag == "while"
    ? t.scope.body
    : t.tag === "if"
        ? t.f
            ? t.t.body.concat(t.f.body)
            : t.t.body
        : undefined;
/**
 * Helper function for {@link walk}. Returns an array of all child nodes for
 * a given term (if any).
 *
 * {@link scopedChildren}
 */
export const allChildren = (t) => scopedChildren(t) ||
    (t.tag === "scope"
        ? t.body
        : t.tag === "ternary"
            ? [t.t, t.f]
            : t.tag === "ret"
                ? [t.val]
                : t.tag === "call" || t.tag === "call_i"
                    ? t.args
                    : t.tag === "sym" && t.init
                        ? [t.init]
                        : t.tag === "decl"
                            ? [t.id]
                            : t.tag === "op1" || t.tag === "swizzle"
                                ? [t.val]
                                : t.tag === "op2"
                                    ? [t.l, t.r]
                                    : t.tag === "assign"
                                        ? [t.r]
                                        : isVec(t) || isMat(t)
                                            ? t.val
                                            : isTerm(t.val)
                                                ? t.val
                                                : undefined);
/**
 * Traverses given AST in depth-first order and applies `visit` and
 * `children` fns to each node. Descends only further if `children`
 * returns an array of child nodes. The `visit` function must accept 2
 * args: the accumulator (`acc`) given to {@link walk} and a tree node. The
 * return value of `visit` becomes the new `acc` value, much like in a
 * reduce operation. {@link walk} itself returns the final `acc`.
 *
 * If `pre` is true (default), the `visit` function will be called prior
 * to visiting a node's children. If false, the visitor is called on the
 * way back up.
 *
 * @param visit -
 * @param children -
 * @param acc -
 * @param tree -
 * @param pre -
 */
export const walk = (visit, children, acc, tree, pre = true) => {
    if (isArray(tree)) {
        tree.forEach((x) => (acc = walk(visit, children, acc, x, pre)));
    }
    else {
        pre && (acc = visit(acc, tree));
        const c = children(tree);
        c && (acc = walk(visit, children, acc, c, pre));
        !pre && (acc = visit(acc, tree));
    }
    return acc;
};
/**
 * Builds dependency graph of given function, by recursively adding all
 * function dependencies. Returns graph.
 *
 * @param fn -
 * @param graph -
 */
export const buildCallGraph = (fn, graph = new DGraph()) => fn.deps && fn.deps.length
    ? fn.deps.reduce((graph, d) => buildCallGraph(d, graph.addDependency(fn, d)), graph)
    : graph.addNode(fn);
export const decl = (id) => ({
    tag: "decl",
    type: id.type,
    id,
});
/**
 * Wraps the given AST node array in `scope` node, optionally as global
 * scope (default false). The interpretation of the global flag is
 * dependent on the target code gen. I.e. for GLSL / JS, the flag
 * disables wrapping the scope's body in `{}`, but else has no
 * difference. In general this node type only serves as internal
 * mechanism for various control flow AST nodes and should not need to
 * be used directly from user land code (though might be useful to
 * create custom / higher level control flow nodes).
 *
 * @param body -
 * @param global -
 */
export const scope = (body, global = false) => ({
    tag: "scope",
    type: "void",
    body: (body
        .filter((x) => x != null)
        .map((x) => (x.tag === "sym" ? decl(x) : x))),
    global,
});
/**
 * Takes an array of global sym/var definitions ({@link input}, {@link output},
 * {@link uniform}) and functions defined via {@link defn}. Constructs the call
 * graph of all transitively used functions and bundles everything in
 * topological order within a global scope object, which is then returned to the
 * user and can be passed to a target codegen for full program output.
 *
 * - {@link scope}
 * - {@link input}
 * - {@link output}
 * - {@link uniform}
 *
 * @param body -
 */
export const program = (body) => {
    const syms = body.filter((x) => x.tag !== "fn");
    const g = body.reduce((acc, x) => (x.tag === "fn" ? buildCallGraph(x, acc) : acc), new DGraph());
    return scope(syms.concat(g.sort()), true);
};
